require 'yaml'
require 'rugged'
require 'artifact/version'

module Artifact

  FORMATS = {
    '.markdown' => "Artifact::MarkdownFile",
    '.html'     => "Artifact::HTMLFile"
  }

  def self.config
    @config ||= Config.new
  end

  def self.configure
    yield config
    unless config.author_name && config.author_email
      raise "Both author_name and author_email are needed."
    end
  end

  def self.root
    config.root
  end

  def self.full_path(path)
    File.join(root, path)
  end

  def self.relative_path(path)
    path.sub(root + '/', '')
  end

  def self.open_file(path)
    load_file(path, ReadableFile)
  end

  def self.new_file(path)
    load_file(path, WritableFile)
  end

  def self.load_file(path, fallback)
    if klass = FORMATS[File.extname(path)]
      Kernel.const_get(klass).new(path)
    else
      fallback.new(path)
    end
  end

  def self.repo
    @repo ||= Repo.new(File.expand_path(config.root))
  end

  def self.ensure_dir!(full_path)
    dir = File.dirname(full_path)
    FileUtils.mkdir_p(dir) unless File.exist?(dir)
  end

  def self.run_hook(name, *args)
    if proc = config.send(name)
      puts " ------------> Running hook: #{name}"
      proc.call(*args)
    end
  end

  class Config

    attr_accessor *%w(
      root source_root posts_path drafts_path uploads_path
      directory_indexes rebuild_args author_name author_email
      before_update after_update before_commit after_commit 
      before_push after_push
    )

    def initialize
      @root              = Dir.pwd # current working path
      @source_root       = 'source'
      @posts_path        = 'posts'
      @drafts_path       = 'drafts'
      @uploads_path      = 'uploads'
      @directory_indexes = true
      @rebuild_args      = ''
    end

  end

  class Tree

    DEFAULT_NEW_FILE_PATH = '%Y/%m'

    attr_reader :path, :full_path

    def initialize(full_path, options = {})
      @full_path = File.expand_path(full_path)
      @path      = Artifact.relative_path(full_path)
      @options   = options

      Artifact.ensure_dir!(full_path)
    end

    def new(filename, extension = nil)
      file_path = File.join(new_file_path, [filename, extension].compact.join)

      if File.exist?(file_path)
        appended = filename.sub(/\-?(\d{1,2})?$/) { |a| "-#{((a || 0).to_i*-1)+1}" }
        return new(appended, extension)
      end

      file_path.sub(filename, filename + '-1')
      Artifact.new_file(file_path)
    end

    def new_file_path
      path = File.join(full_path, @options[:new_file_path] || DEFAULT_NEW_FILE_PATH)
      path.split(File::SEPARATOR).map do |n|
        n['%'] ? Time.now.strftime(n) : n
      end.join(File::SEPARATOR)
    end

    def all
      file_list.map { |f| Artifact.open_file(f) }.sort
    end

    def find(match)
      file_list.select { |f| f[match] }.map { |f| Artifact.open_file(f) }.sort
    end

    def file_list
      file_list_at(full_path)
    end

    def file_list_at(dir, matching = '*')
      files = []
      Dir.glob(File.join(dir, matching)).each do |entry|
        if File.directory?(entry)
          files = files + file_list_at(entry, matching)
        elsif File.file?(entry) # skip sockets and symlinks
          files << entry
        end
      end
      files
    end

  end

  class Repo

    class SyncError < RuntimeError; end

    def initialize(full_path)
      @root = full_path
      @git = Rugged::Repository.new(full_path)
    end

    def save(path, author, check = true)
      # index.add expects a clean path, without any leading ./
      # so make sure we're passing a clean pathname
      path = path.sub(/^\.\//, '')

      if check && !new_file?(path) && !modified?(path)
        raise "No changes in file: #{path}"
      end

      # puts "Adding to index: #{path}"
      @git.index.add(path)
      @git.index.write
      write_commit("Saved #{path}", author)
    end

    def move(from, to, author)
      Artifact.ensure_dir!(Artifact.full_path(to))
      FileUtils.mv(Artifact.full_path(from), Artifact.full_path(to))

      @git.index.remove(from)
      @git.index.add(to)
      @git.index.write
      write_commit("Moved #{from} --> #{to}", author)
    end

    def remove(path, author)
      @git.index.remove(path)
      @git.index.write
      write_commit("Removed #{path}", author)
    end

    def exists?(path)
      !!file_status(path) rescue false
    end

    def sync!
      code, out = pull
      raise SyncError.new("Error when pulling changes: #{out.strip}") unless code == 0
      code, out = push
      raise SyncError.new("Error when pushing changes: #{out.strip}") unless code == 0
    end

    private

    def new_file?(file_path)
      file_status(file_path) == 'new'
    end

    def modified?(file_path)
      file_status(file_path) == 'modified'
    end

    def file_status(file_path)
      if s = @git.status(file_path).first
        return s.to_s.sub('worktree_', '')
      end
    end

    # the author signature is a Hash containing :name, :email of the author
    # and :time of the change.
    def write_commit(message, author)
      Artifact.run_hook(:before_commit, message)

      author[:time] = Time.now
      commit_sha = Rugged::Commit.create(@git,
        :author     => author,
        :message    => message,
        :committer  => author,
        :parents    => @git.empty? ? [] : [ @git.head.target ].compact,
        :tree       => @git.index.write_tree,
        :update_ref => 'HEAD'
      )

      Artifact.run_hook(:after_commit, message)
    end

    def pull
      # updated = false
      code, out = nil, nil
      Dir.chdir(@root) do
        out  = `git pull --rebase`
        code = $? 
        # updated = out.strip != 'Current branch master is up to date.'
        # updated = res.strip != 'Already up-to-date.'
      end
      # updated
      return code, out
    end

    def push
      Artifact.run_hook(:before_push)

      # updated = false
      code, out = nil, nil
      Dir.chdir(@root) do
        out  = `git push`
        code = $?
        # updated = out.strip != 'Everything up-to-date'
      end

      Artifact.run_hook(:after_push)
      # updated
      return code, out
    end

  end

  class ReadableFile

    attr_reader :path, :full_path

    def initialize(path)
      if path[Artifact.root] # path contains root, meaning absolute path
        @path      = Artifact.relative_path(path)
        @full_path = path
      else
        @path      = path
        @full_path = Artifact.full_path(path)
      end
    end

    def filename
      File.basename(path)
    end

    def slug
      File.basename(path, File.extname(path))
    end

    def dirname
      File.dirname(full_path)
    end

    def last_modified
      File.mtime(full_path)
    end

    def body
      @content
    end

    def meta
      {}
    end

    def exists?
      File.exist?(full_path)
    end

    private

    def content
      @content ||= read
    end

    def read
      IO.read(@full_path)
    end

    def <=>(other_file)
      self.last_modified <=> other_file.last_modified
    end

  end

  class WritableFile < ReadableFile

    def save(new_content)
      Artifact.ensure_dir!(@full_path)
      # File.open(@full_path, 'wb') { |f| f.write(content.gsub("\r\n", "\n")) }
      File.open(@full_path, 'wb') { |f| f.write(new_content) }
      @content = nil
    end

    def update(new_content, meta = nil)
      if exists? and new_content == read
        return puts "No changes. No need to write changes to disk."
      end

      Artifact.run_hook :before_update, new_content, @full_path
      save(new_content)
      Artifact.run_hook :after_update, new_content, @full_path
    end

    def delete
      FileUtils.rm(full_path)
    end

  end

  class HTMLFile < WritableFile

  end

  class MarkdownFile < WritableFile

    def title
      data['title']
    end

    def date
      data['date']
    end

    def author
      data['author']
    end

    def category
      data['category']
    end

    def last_updated_by
      data['last_updated_by']
    end

    def update(content, meta = {})
      yaml_data = sanitized_meta(meta) # insert date and author if missing
      body = "#{yaml_data}---\n\n#{content}"
      super(body)
      @data = nil # force reload
    end

    def update_meta(key, val)
      hash = {}
      hash[key.to_s] = val
      update(content, data.merge(hash))
    end

    def content
      parts[1..-1].join("---\n\n")
    end

    def data
      @data ||= YAML.load(parts[0])
    rescue Psych::SyntaxError => e
      raise "Invalid YAML at #{path}: #{parts[0]}"
    end

    def parts
      read.split("---\n\n")
    end

    # include existing date and author in yaml hash in case they're not set
    def sanitized_meta(hash)
      hash.stringify_keys!
      if exists?
        %w(date author).each do |key|
          hash[key] = self.send(key) if !hash[key] and self.send(key)
        end
      end
      hash.to_yaml
    end

  end

end
