require File.join(File.dirname(__FILE__), 'app')
require 'active_support'

require 'middleman-core/sitemap'
require 'middleman-core/core_extensions/front_matter'

class Middleman::Sitemap::Store

  alias_method :original_initialize, :initialize

  def initialize(app)
    Artifact.middleman = app
    original_initialize(app)
  end

end

class Middleman::CoreExtensions::FrontMatter

  # extension's clear_data method doesn't work
  # because it uses the full file path, while
  # the cache setter uses only the relative path
  def flush_cache!
    @cache = {}
  end

end

module Artifact

  def self.middleman=(app)
    @app = app
  end

  def self.middleman
    @app
  end

  def self.source_root
    File.join(config.root, config.source_root)
  end

  def self.posts
    @posts ||= Tree.new(File.join(source_root, Artifact.config.posts_path))
  end

  def self.drafts
    opts = {:new_file_path => ''} # flat, no date
    @drafts ||= Tree.new(File.join(source_root, Artifact.config.drafts_path), opts)
  end

  def self.uploads
    @uploads ||= Tree.new(File.join(source_root, Artifact.config.uploads_path))
  end

  class Middleman < App

    POST_EXTENSION = '.html.markdown'

    get '/' do
      redirect to('/posts')
    end

    post '/build' do
      # updated! and rebuild!
      rebuild!
      redirect to('/')
    end

    post '/sync' do
      begin
        Artifact.repo.sync!
        redirect to('/')
      rescue Artifact::Repo::SyncError => e
        redirect to('/?error=' + e.message)
      end
    end

    get '/search' do
      if params[:q].present?
        normalized = params[:q].gsub(' ', '-').downcase
        @list = filter_and_sort(Artifact.posts.find(normalized), :date)
      else
        @list = filter_and_sort(Artifact.posts.all.last(10), :date)
      end
      partial 'posts/list'
    end

    get '/posts' do
      @posts  = filter_and_sort(Artifact.posts.all.last(10), :date)
      @drafts = filter_and_sort(Artifact.drafts.all, :last_modified)
      view 'posts/index'
    end

    get '/posts/new' do
      view 'posts/new'
    end

    # create new draft (non-existing file)
    post '/posts' do
      data = params[:meta] # includes title and maybe something else
      filename = data[:title].parameterize

      @file = Artifact.drafts.new(filename, POST_EXTENSION)
      @file.update(params[:content], data)

      Artifact.repo.save(@file.path, current_user, false)

      updated!
      redirect to('/posts')
    end

    get '/drafts/*' do
      @post  = MarkdownFile.new(params[:splat][0])
      @draft = true
      view 'posts/edit'
    end

    get '/posts/*' do
      @post = MarkdownFile.new(params[:splat][0])
      view 'posts/edit'
    end

    # save, and optionally publish or unpublish post
    post '/posts/*' do
      @post = MarkdownFile.new(params[:splat][0])

      # ensure date is valid and insert author/last_updated_by
      meta = check_post_meta!(params[:save]) or halt(422, "Invalid meta.")

      # puts "Writing content and meta to file: #{meta.inspect}"
      @post.update(params[:content].gsub("\r\n", "\n"), meta)

      if params[:save] # user clicked on 'save', 'publish' or 'unpublish' (not auto-save)

        # save changes to repo!
        # if draft was previously saved this will raise 'coz no changes.
        Artifact.repo.save(@post.path, current_user, false)

        if params[:save] == 'Publish'
          publish_post!(@post)
        elsif params[:save] == 'Unpublish'
          unpublish_post!(@post)
        else
          @just_saved = true
        end
      end

      updated! # force middleman to rebuild resource list

      if request.xhr?
        "OK"
      else
        redirect @just_saved ? back : to('/posts')
      end
    end

    delete '/posts/*' do
      @post = MarkdownFile.new(params[:splat][0])
      @post.delete
      Artifact.repo.remove(@post.path, current_user) if Artifact.repo.exists?(@post.path)
      updated!

      redirect to('/posts')
    end

    get '/uploads' do
      get_uploads
      # @uploads.select! {|i| i.match(/\.(png|jpg|jpeg|gif)/i) }
      partial 'uploads/list'
    end

    post '/uploads' do
      unless params[:file] &&
        (tmpfile = params[:file][:tempfile]) &&
        (name = params[:file][:filename])
        raise 'No file uploaded.'
      else
        filename = File.basename(name, File.extname(name)).parameterize # without extension
        file = Artifact.uploads.new(filename, File.extname(name))
        file.save(tmpfile.read)
        Artifact.repo.save(file.path, current_user) # pass relative path

        updated! # force middleman to rebuild resource list
        redirect to('/')
      end
    end

    private

    def check_post_meta!(action)
      meta = (params[:meta] || {}).symbolize_keys.merge(:last_updated_by => current_user[:email])

      if !meta[:date] or action.to_s.downcase.to_sym == :publish
        meta[:date] = Time.now.utc.strftime("%Y-%m-%d %H:%M %Z")
      else
        # puts "Meta date is #{meta[:date]}"
        return false unless valid_date?(meta[:date])
      end

      meta[:author] = current_user[:name] unless meta[:author]
      meta
    end

    def valid_date?(str)
      Time.parse(str) rescue false
    end

    def get_uploads
      @uploads = Artifact.uploads.all.last(10).reverse
    end

    def filter_and_sort(posts, criteria)
      posts.select { |f| f.is_a?(WritableFile) }.sort do |a,b|
        b.send(criteria) <=> a.send(criteria)
      end
    end

    def publish_post!(post)
      filename = post.title.parameterize
      new_file = Artifact.posts.new(filename, POST_EXTENSION)
      Artifact.repo.move(post.path, new_file.path, current_user)
    end

    def unpublish_post!(post)
      filename = post.title.parameterize
      new_file = Artifact.drafts.new(filename, POST_EXTENSION)
      Artifact.repo.move(post.path, new_file.path, current_user)
    end

    def rebuild!
      puts `bundle exec middleman build #{Artifact.config.rebuild_args}`
    end

    def updated!
      return if Artifact.middleman.nil? # not loaded yet!

      # set environment to build so the whole app doesn't get reloaded
      # for each file inside the reload_path block
      # ref: https://github.com/middleman/middleman/blob/v3-stable/middleman-core/lib/middleman-core/sitemap/extensions/on_disk.rb#L62
      Artifact.middleman.config[:environment] = :build

      puts "Reloading Middleman..."
      Artifact.middleman.files.reload_path(Artifact.drafts.path)
      Artifact.middleman.files.reload_path(Artifact.posts.path)
      Artifact.middleman.files.reload_path(Artifact.uploads.path)

      Artifact.middleman.cache.clear # Tilt (template) cache
      # Artifact.middleman.sitemap.send(:reset_lookup_cache!)
      Artifact.middleman.sitemap.rebuild_resource_list!

      if Artifact.middleman.extensions[:frontmatter]
        # puts "Flushing frontmatter cache..."
        Artifact.middleman.extensions[:frontmatter].flush_cache!
      end
    end

  end

end
